use gag::BufferRedirect;
use std::{
    error::Error,
    ffi::{c_char, CString},
    io::Read,
};

use wmenu_sys::bindings;

unsafe extern "C" fn print_item(menu: *mut bindings::menu, text: *mut i8, exit: bool) {
    wmenu_sys::bindings::puts(text);
    wmenu_sys::bindings::fflush(bindings::stdout);

    if exit {
        (*menu).exit = true;
    }
}

fn to_c_char(text: &str) -> *mut c_char {
    return CString::new(text).unwrap().into_raw();
}

#[derive(Debug)]
pub struct MenuError {
    message: String,
    status: i32,
}

impl MenuError {
    pub fn new(status: i32, message: String) -> Self {
        Self { message, status }
    }
}

impl std::fmt::Display for MenuError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "exit code {}: {}", self.status, self.message)
    }
}

impl Error for MenuError {}

pub struct Menu {
    ctx: *mut bindings::menu,
}

impl Menu {
    pub fn new() -> Menu {
        Self {
            ctx: unsafe { bindings::menu_create(Some(print_item)) },
        }
    }

    fn into_inner(&mut self) -> &mut bindings::menu {
        unsafe { &mut *self.ctx }
    }

    pub fn bottom(&mut self, value: bool) -> &mut Self {
        self.into_inner().bottom = value;
        self
    }

    pub fn password(&mut self, value: bool) -> &mut Self {
        self.into_inner().passwd = value;
        self
    }

    pub fn font(&mut self, value: &str) -> &mut Self {
        self.into_inner().font = to_c_char(value);
        self
    }

    pub fn lines(&mut self, value: i32) -> &mut Self {
        self.into_inner().lines = value;
        self
    }

    pub fn output_name(&mut self, value: &str) -> &mut Self {
        self.into_inner().output_name = to_c_char(value);
        self
    }

    pub fn prompt(&mut self, value: &str) -> &mut Self {
        self.into_inner().prompt = to_c_char(value);
        self
    }

    pub fn normal_bg(&mut self, value: u32) -> &mut Self {
        self.into_inner().normalbg = value;
        self
    }

    pub fn normal_fg(&mut self, value: u32) -> &mut Self {
        self.into_inner().normalfg = value;
        self
    }

    pub fn prompt_fg(&mut self, value: u32) -> &mut Self {
        self.into_inner().promptfg = value;
        self
    }

    pub fn prompt_bg(&mut self, value: u32) -> &mut Self {
        self.into_inner().promptbg = value;
        self
    }

    pub fn selection_bg(&mut self, value: u32) -> &mut Self {
        self.into_inner().selectionbg = value;
        self
    }

    pub fn selection_fg(&mut self, value: u32) -> &mut Self {
        self.into_inner().selectionbg = value;
        self
    }

    pub fn width(&mut self, value: i32) -> &mut Self {
        self.into_inner().width = value;
        self
    }

    pub fn height(&mut self, value: i32) -> &mut Self {
        self.into_inner().height = value;
        self
    }

    pub fn line_height(&mut self, value: i32) -> &mut Self {
        self.into_inner().line_height = value;
        self
    }

    pub fn padding(&mut self, value: i32) -> &mut Self {
        self.into_inner().padding = value;
        self
    }

    pub fn add_item(&mut self, item: &str) -> &mut Self {
        unsafe {
            bindings::menu_add_item(self.ctx, to_c_char(item));
        }

        self
    }

    pub fn add_items(&mut self, items: &[&str]) -> &mut Self {
        for item in items {
            self.add_item(item);
        }

        self
    }

    pub fn run(&self) -> Result<String, MenuError> {
        unsafe {
            let mut stdout = BufferRedirect::stdout().unwrap();
            match bindings::menu_run(self.ctx) {
                0 => {
                    let mut buf = String::new();
                    stdout.read_to_string(&mut buf).unwrap();
                    Ok(buf.trim_end().to_string())
                }
                n => {
                    let mut buf = String::new();
                    stdout.read_to_string(&mut buf).unwrap();
                    Err(MenuError::new(n, buf))
                }
            }
        }
    }
}

impl Drop for Menu {
    fn drop(&mut self) {
        unsafe {
            bindings::menu_destroy(self.ctx);
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn awawa() {
        let items = ["roewrn", "sybil", "vex"];

        let mut menu = Menu::new();
        let menu = menu
            .font("monospace 12")
            .height(32)
            .line_height(32)
            .padding(8)
            .add_items(&items);

        let result = menu.run().expect("couldn't get result");
        assert!(items.contains(&result.as_str()));
    }
}
